package eu.mhsl.minecraft.pixelblocks;

import eu.mhsl.minecraft.pixelblocks.pixelblock.PixelBlock;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import static eu.mhsl.minecraft.pixelblocks.PixelBlocks.plugin;

public class DataBase {
    public static String url;

    public DataBase(String url) {
        DataBase.url = url;
    }

    public Connection getConnection() throws SQLException {
        return DriverManager.getConnection(url);
    }

    public Statement getStatement() throws SQLException {
        return getConnection().createStatement();
    }

    public void removePixelBlock(PixelBlock pixelBlock) {
        Bukkit.getScheduler().runTask(plugin, () -> {

            try (var conn = getConnection()) {
                PreparedStatement prep = conn.prepareStatement("DELETE FROM pixelblocks WHERE uuid = ?");
                prep.setString(1, pixelBlock.uuid.toString());
                prep.executeUpdate();

                prep.close();
            } catch (SQLException e) {
                System.err.println(e.getMessage());
            }

        });
    }

    public void savePixelBlock(PixelBlock pixelBlock) {
        Bukkit.getScheduler().runTask(plugin, () -> {
            List<UUID> uuids = new ArrayList<>();

            try (var statement = getStatement()) {
                ResultSet pixelBlocksResult = statement.executeQuery("SELECT * FROM pixelblocks;");
                while (pixelBlocksResult.next()) {
                    uuids.add(UUID.fromString(pixelBlocksResult.getString("uuid")));
                }
            } catch (SQLException e) {
                System.err.println(e.getMessage());
            }

            if(!uuids.contains(pixelBlock.uuid)) {

                try (var conn = getConnection()) {
                    PreparedStatement prep = conn.prepareStatement(
                            "INSERT INTO pixelblocks(uuid, owner, " +
                                    "locationWorldName, locationX, locationY, locationZ, " +
                                    "entryLocationWorldName, entryLocationX, entryLocationY, entryLocationZ) " +
                                    "VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?);"
                    );
                    prep.setString(1, pixelBlock.uuid.toString());
                    prep.setString(2, pixelBlock.ownerUID.toString());

                    prep.setString(3, pixelBlock.pixelBlockLocation.getWorld().getName());
                    prep.setDouble(4, pixelBlock.pixelBlockLocation.getX());
                    prep.setDouble(5, pixelBlock.pixelBlockLocation.getY());
                    prep.setDouble(6, pixelBlock.pixelBlockLocation.getZ());

                    if(pixelBlock.lastEntryLocation != null) {
                        prep.setString(7, pixelBlock.lastEntryLocation.getWorld().getName());
                        prep.setDouble(8, pixelBlock.lastEntryLocation.getX());
                        prep.setDouble(9, pixelBlock.lastEntryLocation.getY());
                        prep.setDouble(10, pixelBlock.lastEntryLocation.getZ());
                    } else {
                        prep.setString(7, Bukkit.getWorlds().getFirst().getName());
                        prep.setDouble(8, Bukkit.getWorlds().getFirst().getSpawnLocation().getX());
                        prep.setDouble(9, Bukkit.getWorlds().getFirst().getSpawnLocation().getY());
                        prep.setDouble(10, Bukkit.getWorlds().getFirst().getSpawnLocation().getZ());
                    }

                    prep.executeUpdate();
                    prep.close();
                } catch (SQLException e) {
                    System.err.println(e.getMessage());
                }

            } else {
                try (var conn = getConnection()) {
                    PreparedStatement prep = conn.prepareStatement(
                            "UPDATE pixelblocks " +
                                    "SET owner=?, locationWorldName=?, locationX=?, locationY=?, locationZ=?," +
                                    "entryLocationWorldName=?, entryLocationX=?, entryLocationY=?, entryLocationZ=? " +
                                    "WHERE uuid=?;"
                    );
                    prep.setString(1, pixelBlock.ownerUID.toString());

                    prep.setString(2, pixelBlock.pixelBlockLocation.getWorld().getName());
                    prep.setDouble(3, pixelBlock.pixelBlockLocation.getX());
                    prep.setDouble(4, pixelBlock.pixelBlockLocation.getY());
                    prep.setDouble(5, pixelBlock.pixelBlockLocation.getZ());

                    if(pixelBlock.lastEntryLocation != null) {
                        prep.setString(6, pixelBlock.lastEntryLocation.getWorld().getName());
                        prep.setDouble(7, pixelBlock.lastEntryLocation.getX());
                        prep.setDouble(8, pixelBlock.lastEntryLocation.getY());
                        prep.setDouble(9, pixelBlock.lastEntryLocation.getZ());
                    } else {
                        prep.setString(6, Bukkit.getWorlds().getFirst().getName());
                        prep.setDouble(7, Bukkit.getWorlds().getFirst().getSpawnLocation().getX());
                        prep.setDouble(8, Bukkit.getWorlds().getFirst().getSpawnLocation().getY());
                        prep.setDouble(9, Bukkit.getWorlds().getFirst().getSpawnLocation().getZ());
                    }

                    prep.setString(10, pixelBlock.uuid.toString());

                    prep.executeUpdate();
                    prep.close();
                } catch (SQLException e) {
                    System.err.println(e.getMessage());
                }
            }
        });
    }

    public void loadPixelBlocks() {
        List<Entity> entities = new ArrayList<>();
        entities.addAll(Bukkit.getWorlds().get(0).getEntities().stream()
                .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                .toList());
        entities.addAll(Bukkit.getWorlds().get(1).getEntities().stream()
                .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                .toList());
        entities.addAll(Bukkit.getWorlds().get(2).getEntities().stream()
                .filter(entity -> entity.getType().equals(EntityType.BLOCK_DISPLAY) || entity.getType().equals(EntityType.INTERACTION) || entity.getType().equals(EntityType.ITEM_DISPLAY))
                .toList());

        String sql = "CREATE TABLE IF NOT EXISTS pixelblocks (" +
                "uuid CHAR(36) PRIMARY KEY, " +
                "owner CHAR(36), " +
                "locationWorldName CHAR(36), " +
                "locationX DOUBLE, " +
                "locationY DOUBLE, " +
                "locationZ DOUBLE, " +
                "entryLocationWorldName CHAR(36), " +
                "entryLocationX DOUBLE, " +
                "entryLocationY DOUBLE, " +
                "entryLocationZ DOUBLE" +
                ");";

        try (var statement = getStatement()) {
            statement.execute(sql);
            ResultSet pixelBlocksResult = statement.executeQuery("SELECT * FROM pixelblocks;");

            while (pixelBlocksResult.next()) {
                Location newPixelBlockLocation = new Location(
                        Bukkit.getWorld(pixelBlocksResult.getString("locationWorldName")),
                        pixelBlocksResult.getDouble("locationX"),
                        pixelBlocksResult.getDouble("locationY"),
                        pixelBlocksResult.getDouble("locationZ")
                );

                Location newEntryLocation = new Location(
                        Bukkit.getWorld(pixelBlocksResult.getString("entryLocationWorldName")),
                        pixelBlocksResult.getDouble("entryLocationX"),
                        pixelBlocksResult.getDouble("entryLocationY"),
                        pixelBlocksResult.getDouble("entryLocationZ")
                );

                entities.stream().filter(entity -> entity.
                        getLocation().
                        clone().
                        add(0, PixelBlock.hitboxOffset, 0).
                        toBlockLocation().
                        equals(newPixelBlockLocation)).
                        forEach(Entity::remove);

                PixelBlock newPixelBlock = new PixelBlock(
                        newPixelBlockLocation,
                        UUID.fromString(pixelBlocksResult.getString("owner")),
                        UUID.fromString(pixelBlocksResult.getString("uuid"))
                );
                newPixelBlock.lastEntryLocation = newEntryLocation;
                newPixelBlock.place(newPixelBlockLocation);
            }

        } catch (SQLException e) {
            System.err.println(e.getMessage());
        }

    }
}
